package de.unipotsdam.hpi.cind.associationRuleMining;

import java.util.*;

import de.unipotsdam.hpi.cind.associationRuleMining.util.*;

public class AprioriConstrained extends Apriori {

	protected ArrayList<Basket> nonIncludedBaskets;
	
	public AprioriConstrained(Collection<Basket> baskets, 
			int minSupport, double minValidity, double minTotalCompleteness) {
		
		super(baskets, minSupport, minValidity, minTotalCompleteness);
	}

	public AprioriConstrained(Collection<Basket> includedBaskets, Collection<Basket> nonIncludedBaskets, 
			int minSupport, double minValidity, double minTotalCompleteness) {
		
		this(includedBaskets, minSupport, minValidity, minTotalCompleteness);
		this.nonIncludedBaskets = new ArrayList<Basket>(nonIncludedBaskets);
	}
	
	
	/**
	 * Build two item sets by combining each frequent one item set with item
	 * "INCLUDED_TRUE". Remove infrequent item sets.
	 * 
	 * @return frequent two item sets
	 */
	@Override
	protected HashTreeNode determineTwoItemSets(ArrayList<ItemSetString> frequentOneItemSets) {
		
		int splitSize = Math.min(HashTreeNode.DEFAULT_SPLIT_THRESHOLD, frequentOneItemSets.size()/3);
		HashTreeNode twoItemSets = new HashTreeNode(splitSize);
//		System.err.println("splitSize (AprioriConstraint.twoItemSets): " + splitSize);
		
		/*
		 * Combine each frequent one item set with item
		 * "INCLUDED_TRUE".
		 */
		for (ItemSetString itemSetString : frequentOneItemSets) {
			if (itemSetString.getItem(0).equals(Apriori.INCLUDED_TRUE))
				continue;
			
			twoItemSets.addItemSet(new ItemSetString(Apriori.INCLUDED_TRUE, itemSetString.getItem(0)));
		}
		
		return twoItemSets;
		
	}

	/**
	 * Count item set frequencies of specified item sets.
	 * 
	 * 
	 * @param kItemSets
	 * @return true, if at least one basket was deleted (i.e., always false as
	 *         constrained version needs all baskets for rule generation and
	 *         thus, no baskets are deleted)
	 */
	@Override
	protected boolean countItemSetFrequencies(HashTreeNode kItemSets) {
		
		for (Basket currentBasket : baskets) {
			kItemSets.increaseSupportOfContainedItemSets(currentBasket);
		}
		
		// constrained version needs all baskets for rule generation
		// thus, no baskets deleted 
		return false;
	}
	
	/**
	 * Count item set frequencies of specified item sets in *all* baskets, i.e.
	 * included and non-included tuples.
	 * 
	 * 
	 * @param kItemSets
	 * @return true, if at least one basket was deleted (i.e., always false as
	 *         constrained version needs all baskets for rule generation and
	 *         thus, no baskets are deleted)
	 */
	protected boolean countItemSetFrequenciesAllBaskets(HashTreeNode kItemSets) {
		
		for (Basket currentBasket : baskets) {
			kItemSets.increaseSupportOfContainedItemSets(currentBasket);
		}
		
		for (Basket currentBasket : nonIncludedBaskets) {
			kItemSets.increaseSupportOfContainedItemSets(currentBasket);
		}
		
		// constrained version needs all baskets for rule generation
		// thus, no baskets deleted 
		return false;
	}
	
	/**
	 * Test all sub item sets except i_1,...,i_k-1 (= itemSetPrefix1) and
	 * i_1,...,i_k-2,i_k (= itemSetPrefix2), but first item is never omitted,
	 * i.e., only sub sets with {@link INCLUDED_TRUE},
	 * 
	 * and
	 * 
	 * look for sub item sets in frequent item sets and omit candidate if sub
	 * item set is not contained
	 * 
	 * @param frequentItemSets
	 *            hash tree with item sets of size itemSetSize
	 * @param itemSetSize
	 * @param candidate
	 *            item set of size itemSetSize+1
	 * 
	 * @return true, if all item sub sets are frequent
	 * 
	 */
	@Override
	protected boolean allSubSetsFrequent(
			HashTreeNode frequentItemSets, int itemSetSize, ItemSetString candidate) {
		
		List<String> completeItemSet = candidate.getItems();
		ItemSetString currentSubItemSet;
		for (int i = 1; i < itemSetSize-1; i++) {
			// get sub item set
			currentSubItemSet = new ItemSetString(
					completeItemSet.subList(0, i), 
					completeItemSet.subList(i+1, itemSetSize+1));

			// test its frequency
			if ( !frequentItemSets.contains(currentSubItemSet) ) {
				// found infrequent sub set
				return false;
			}
		}

		// all sub item sets contained -> valid candidate
		return true;
	}

	
	/**
	 * 
	 * @param frequentItemSets
	 */
	@Override
	public ArrayList<AssociationRule> generateAssociationRulesWithRhsTrue(
			HashMap<List<String>, Integer> frequentItemSets) {
		

		ArrayList<AssociationRule> result = 
			new ArrayList<AssociationRule>(frequentItemSets.size());
		
		
		/*
		 * get frequencies of frequent item sets without INCLUDED_TRUE
		 */
		// get lhs item sets
		ArrayList<ItemSetString> lhsItemSets = 
			new ArrayList<ItemSetString>(frequentItemSets.size());
		
		for (List<String> currentItemSet : frequentItemSets.keySet()) {
			if (currentItemSet.size() < 2)
				continue;
			
			List<String> lhs = currentItemSet.subList(1, currentItemSet.size());
			lhsItemSets.add(new ItemSetString(lhs));
		}
		
		// count lhs item set frequencies
//		int splitSize = lhsItemSets.size()/3;
		int splitSize = Math.min(
				HashTreeNode.DEFAULT_SPLIT_THRESHOLD, lhsItemSets.size()/largestItemSetSize);
		HashTreeNode lhsItemSetsHashTree = new HashTreeNode(splitSize);
		for (ItemSetString lhsItemSet : lhsItemSets) {
			lhsItemSetsHashTree.addItemSet(lhsItemSet);
		}
		countItemSetFrequenciesAllBaskets(lhsItemSetsHashTree);
		
//		System.err.println("frequentItemSets.size(): " + frequentItemSets.size());
//		System.err.println("splitSize (AprioriConstraint.ruleGen): " + splitSize);
//		System.err.println("lhsItemSets.size(): " + lhsItemSets.size());
		
		// save item sets and frequencies as map
		HashMap<List<String>, Integer> lhsItemSetsMap = new HashMap<List<String>, Integer>(frequentItemSets.size());
		for (ItemSetString currentItemSet : lhsItemSets) {
			lhsItemSetsMap.put(currentItemSet.getItems(), currentItemSet.getFrequency());
		}
		

		/*
		 * confidence = validity = support(X \cup INCLUDED_TRUE) / support(X)
		 * confidenceBar = totalCompleteness = support(X \cup INCLUDED_TRUE) / support(INCLUDED_TRUE)
		 */
		
		// support(INCLUDED_TRUE)
//		int includedTrueSupp = frequentItemSets.get(Arrays.asList(INCLUDED_TRUE));
		// support(X \cup INCLUDED_TRUE)
		int lhsRhsSupp;
		// support(X)
		int lhsSupp ;
		
		double validity, totalCompleteness;
		
		for (List<String> currentItemSet : frequentItemSets.keySet()) {
			
			if (currentItemSet.size() > 1) {
				
				lhsRhsSupp = frequentItemSets.get(currentItemSet);
				
				List<String> lhs = currentItemSet.subList(1, currentItemSet.size());
				lhsSupp = lhsItemSetsMap.get(lhs);
				if (lhsSupp == 0)
					throw new RuntimeException(lhs + "not contained in lhsItemSetsMap");
//					lhsSupp = frequentItemSets.get(lhs);
				
				validity = (double) lhsRhsSupp / lhsSupp;
				totalCompleteness = (double) lhsRhsSupp / includedTrueSupp; 
				
				if (validity >= minValidity && totalCompleteness >= minTotalCompleteness) {
					result.add(new AssociationRule(
							lhs, lhsRhsSupp, lhsSupp, validity, totalCompleteness));
				}
			}
			
		}
		
		return result;
	}
	
}
