package de.unipotsdam.hpi.cind.associationRuleMining;

import java.sql.*;
import java.util.*;

import de.unipotsdam.hpi.cind.associationRuleMining.util.*;


public class CindDetectionUsingApriori {

	final static String[] PREFIXES = { "A", "B", "C", "D", "E", "F",
			"G", "H", "I", "J", "K", "L", "M", "N", "O", "P", "Q",
			"R", "S", "T", "U", "V", "W", "X", "Y", "Z" };

	protected Connection conn;
	
	public CindDetectionUsingApriori(Connection conn) {
		this.conn = conn;
	}

	/**
	 * Executes and reads data from specified query as baskets. Each value is
	 * preceded by "<attribute number>|" to preserve column affiliation. Calls
	 * Apriori to find CINDs.
	 * 
	 * @param query
	 * @param numberOfAttributes
	 * @param minSupport
	 * @throws SQLException
	 */
	public ArrayList<ConditionalIND> findConditionalINDsApriori(
			String query, String[] attributeNames, 
			double minSupport, double minValidity, double minTotalCompleteness) 
			throws SQLException {
		
		
		/*
		 * get prefixes
		 */
		String[] prefixes = getPrefixes(attributeNames.length);
    	HashMap<String, String> prefixesToAttributeNames = new HashMap<String, String>(attributeNames.length * 2);
    	for (int i = 0; i < prefixes.length; i++) {
			prefixesToAttributeNames.put(prefixes[i], attributeNames[i]);
		}
    	
    	
    	/*
    	 * read baskets
    	 */
		ArrayList<Basket> baskets = readBaskets( query, prefixes );
		
		
		/*
		 * call Apriori
		 */
		Apriori apriori = new Apriori(
				baskets, 
				(int) (minSupport * baskets.size()),
				minValidity,
				minTotalCompleteness);
    	
    	HashMap<List<String>, Integer> frequentItemSets = apriori.findFrequentItemSets();
    	ArrayList<AssociationRule> rules = apriori.generateAssociationRulesWithRhsTrue(frequentItemSets);
    	
    	
    	/*
    	 * create CINDs from rules
    	 */
    	ArrayList<ConditionalIND> cinds = new ArrayList<ConditionalIND>(rules.size());
    	for (AssociationRule currentRule : rules) {
			cinds.add( new ConditionalIND(
					currentRule, 
					prefixesToAttributeNames, 
					prefixes[0].length(), 
					baskets.size()) );;
		}
    	
    	return cinds;
	}
	

	/**
	 * Executes and reads data from specified query as baskets. Each value is
	 * preceded by "<attribute number>|" to preserve column affiliation. Calls
	 * AprioriConstrained to find CINDs.
	 * 
	 * @param query
	 * @param numberOfAttributes
	 * @param minSupport
	 * @return
	 * @throws SQLException
	 */
	public ArrayList<ConditionalIND> findConditionalINDsAprioriConstrained(
			String query, String[] attributeNames, 
			double minSupport, double minValidity, double minTotalCompleteness) 
			throws SQLException {
		
//		System.err.println(Arrays.toString(attributeNames));
		
		/*
		 * get prefixes 
		 */
		String[] prefixes = getPrefixes(attributeNames.length);
    	HashMap<String, String> prefixesToAttributeNames = 
    		new HashMap<String, String>(attributeNames.length * 2);
    	for (int i = 0; i < prefixes.length; i++) {
			prefixesToAttributeNames.put(prefixes[i], attributeNames[i]);
		}

    	/*
    	 * read baskets 
    	 */
		List<ArrayList<Basket>> baskets = readBasketsConstrained(query, prefixes);
		ArrayList<Basket> includedBaskets = baskets.get(0);
		ArrayList<Basket> nonIncludedBaskets = baskets.get(1);
		
		/*
		 *  call AprioriConstrained
		 */
		AprioriConstrained apriori = new AprioriConstrained(
				includedBaskets, nonIncludedBaskets,  
				(int) (minSupport * (includedBaskets.size() + nonIncludedBaskets.size())),
				minValidity,
				minTotalCompleteness);
    	
    	HashMap<List<String>, Integer> frequentItemSets = 
    		apriori.findFrequentItemSets();
    	

    	long startRuleGen = System.currentTimeMillis();
    	
    	ArrayList<AssociationRule> rules = 
    		apriori.generateAssociationRulesWithRhsTrue(frequentItemSets);

    	long end = System.currentTimeMillis();

    	/*
    	 * create CINDs from rules
    	 */
    	ArrayList<ConditionalIND> cinds = new ArrayList<ConditionalIND>(rules.size());
    	for (AssociationRule currentRule : rules) {
			cinds.add( new ConditionalIND(
					currentRule, 
					prefixesToAttributeNames, 
					prefixes[0].length(), 
					(includedBaskets.size() + nonIncludedBaskets.size())) );;
		}
    	
    	
    	RuntimeMeasures runtimes = RuntimeMeasures.getInstance();
    	runtimes.setRuntimeRuleGen( (int)(end-startRuleGen) );
    	
    	return cinds;
    	
	}

	/**
	 * Executes and reads data from specified query as baskets. Each value is
	 * preceded by "<attribute number>|" to preserve column affiliation. Calls
	 * AprioriConstrained to find CINDs. 
	 * 
	 * @param query
	 * @param numberOfAttributes
	 * @param minSupportAbsolute
	 * @return
	 * @throws SQLException
	 */
	public ArrayList<ConditionalIND> findConditionalINDsAprioriConstrained(
			String query, String[] attributeNames, 
			int minSupportAbsolute, double minValidity, double minTotalCompleteness) 
			throws SQLException {
		
		
		/*
		 * get prefixes 
		 */
		String[] prefixes = getPrefixes(attributeNames.length);
    	HashMap<String, String> prefixesToAttributeNames = 
    		new HashMap<String, String>(attributeNames.length * 2);
    	for (int i = 0; i < prefixes.length; i++) {
			prefixesToAttributeNames.put(prefixes[i], attributeNames[i]);
		}

    	/*
    	 * read baskets 
    	 */
		List<ArrayList<Basket>> baskets = readBasketsConstrained(query, prefixes);
		ArrayList<Basket> includedBaskets = baskets.get(0);
		ArrayList<Basket> nonIncludedBaskets = baskets.get(1);
		
		
		/*
		 *  call AprioriConstrained
		 */
		AprioriConstrained apriori = new AprioriConstrained(
				includedBaskets, nonIncludedBaskets, 
				minSupportAbsolute,
				minValidity,
				minTotalCompleteness);
    	
		
    	HashMap<List<String>, Integer> frequentItemSets = 
    		apriori.findFrequentItemSets();
    	

    	long startRuleGen = System.currentTimeMillis();
    	
    	ArrayList<AssociationRule> rules = 
    		apriori.generateAssociationRulesWithRhsTrue(frequentItemSets);

    	long end = System.currentTimeMillis();

    	/*
    	 * create CINDs from rules
    	 */
    	ArrayList<ConditionalIND> cinds = new ArrayList<ConditionalIND>(rules.size());
    	for (AssociationRule currentRule : rules) {
			cinds.add( new ConditionalIND(
					currentRule, 
					prefixesToAttributeNames, 
					prefixes[0].length(), 
					baskets.size()) );;
		}
    	
    	
    	RuntimeMeasures runtimes = RuntimeMeasures.getInstance();
    	runtimes.setRuntimeRuleGen( (int)(end-startRuleGen));
    	
    	return cinds;
    	
	}

	
	
	/**
	 * 
	 * 
	 * @param attributeNames
	 * @return prefixes of fixed length: if nrOfAttributes < PREFIXES.length
	 *         than length of prefixes will be one, if PREFIXES.length <
	 *         nrOfAttributes < PREFIXES.length^2 than length of prefixes will
	 *         be two, otherwise an Exception is thrown
	 */
	private String[] getPrefixes(int nrOfAttributes) throws RuntimeException {


		String[] prefixes;
		if (nrOfAttributes < PREFIXES.length)
			prefixes = Arrays.copyOf(PREFIXES, nrOfAttributes);
		else if (nrOfAttributes < (PREFIXES.length * PREFIXES.length)){
			prefixes = new String[nrOfAttributes];
			prefixes = Arrays.copyOf(PREFIXES, nrOfAttributes);
			for (int i = PREFIXES.length; i < prefixes.length; i++) {
				prefixes[i] = PREFIXES[i/PREFIXES.length -1] + PREFIXES[i % PREFIXES.length];
			}
		} else {
			throw new RuntimeException("Unable to create prefixes for "
					+ nrOfAttributes
					+ " attributes. Maximum number of prefixes is "
					+ (PREFIXES.length * PREFIXES.length)
					+ ". Add prefixes to CindDetectionUsingApriori.PREFIXES.");
		}
		
		return prefixes;
	}

	/**
	 * Executes specified query and reads data into baskets. Each value is
	 * preceded by a prefix to preserve column affiliation.
	 * 
	 * 
	 * @param query
	 *            Query to get outer join over depending and referenced
	 *            relations. First attribute is referenced column indicating
	 *            inclusion (NOT NULL) or not (NULL).
	 * @param prefixes
	 *            Prefixes for dependent columns, i.e., column
	 *            2,...,prefixes.length+1
	 * @return
	 * @throws SQLException
	 */
	protected ArrayList<Basket> readBaskets(
			String query, String[] prefixes) 
			throws SQLException {
		
		long start = System.currentTimeMillis();
		
		
		// get data as left outer join 
		Statement stmt = conn.createStatement();;
		ResultSet rsBaskets = stmt.executeQuery(query);
		
		// store data to baskets with
		// - referenced column 0 or 1 
		// - values of first column prefixed with 1|, second column with 2|, ...
		ArrayList<Basket> baskets = new ArrayList<Basket>(1024);
		ArrayList<String> currentBasket = new ArrayList<String>(prefixes.length);
		Object currentItem;
		
		while( rsBaskets.next() ) {
			
			// handle INCLUDED_TRUE, INCLUDED_FALSE
			if (rsBaskets.getObject(1) == null)
				currentBasket.add(Apriori.INCLUDED_FALSE);
			else
				currentBasket.add(Apriori.INCLUDED_TRUE);
			
			// handle other attributes
			for (int i = 2; i <= prefixes.length+1; i++) {
				
				currentItem = rsBaskets.getObject(i);
				if (currentItem == null )
					continue;
				
				currentBasket.add(prefixes[i-2] + currentItem.toString());
			}
			
			baskets.add( new Basket(currentBasket) );
			currentBasket.clear();
			
		}
		
		rsBaskets.close();
		stmt.close();
		
		long end = System.currentTimeMillis();
		
		RuntimeMeasures runtimes = RuntimeMeasures.getInstance();
		runtimes.setRuntimeReadData( (int)(end-start) );

		return baskets;
	}
	
	/**
	 * Executes specified query and reads data into baskets. Each value is
	 * preceded by a prefix to preserve column affiliation.
	 * 
	 * 
	 * @param query
	 *            Query to get outer join over depending and referenced
	 *            relations. First attribute is referenced column indicating
	 *            inclusion (NOT NULL) or not (NULL).
	 * @param prefixes
	 *            Prefixes for dependent columns, i.e., column
	 *            2,...,prefixes.length+1
	 * @return
	 * @throws SQLException
	 */
	protected List<ArrayList<Basket>> readBasketsConstrained(
			String query, String[] prefixes) 
			throws SQLException {
		
		long start = System.currentTimeMillis();
		
		
		// get data as left outer join 
		Statement stmt = conn.createStatement();;
		ResultSet rsBaskets = stmt.executeQuery(query);
		
		// store data to baskets with
		// - referenced column 0 or 1 
		// - values of first column prefixed with 1|, second column with 2|, ...
		ArrayList<Basket> includedBaskets = new ArrayList<Basket>(1024);
		ArrayList<Basket> nonIncludedBaskets = new ArrayList<Basket>(1024);
		boolean isIncluded = false;
		
		ArrayList<String> currentBasket = new ArrayList<String>(prefixes.length);
		Object currentItem;
		
		
		while( rsBaskets.next() ) {
			
			// handle INCLUDED_TRUE, INCLUDED_FALSE
			if (rsBaskets.getObject(1) == null) {
				currentBasket.add(Apriori.INCLUDED_FALSE);
				isIncluded = false;
			}
			else {
				currentBasket.add(Apriori.INCLUDED_TRUE);
				isIncluded = true;
			}
			
			// handle other attributes
			for (int i = 2; i <= prefixes.length+1; i++) {
				
				currentItem = rsBaskets.getObject(i);
				if (currentItem == null )
					continue;
				
				currentBasket.add(prefixes[i-2] + currentItem.toString());
			}
			
			if (isIncluded)
				includedBaskets.add( new Basket(currentBasket) );
			else
				nonIncludedBaskets.add( new Basket(currentBasket) );
			currentBasket.clear();
			
		}
		
		rsBaskets.close();
		stmt.close();
		
		long end = System.currentTimeMillis();
		
		RuntimeMeasures runtimes = RuntimeMeasures.getInstance();
		runtimes.setRuntimeReadData( (int)(end-start) );

		ArrayList<ArrayList<Basket>> baskets = new ArrayList<ArrayList<Basket>>();
		baskets.add(includedBaskets);
		baskets.add(nonIncludedBaskets);
		
		return baskets;
	}
	
	
}
