package de.unipotsdam.hpi.cind.associationRuleMining;

import java.io.*;
import java.text.*;
import java.util.*;

import java.sql.*;        


import de.hpi.fgis.dude.util.*;
import de.unipotsdam.hpi.cind.associationRuleMining.util.*;


public class CindDetectionUsingAprioriExec {

	/**
	 * @param args
	 *            <ul>
	 *            <li>-supp &lt; minSupport &gt;</li>
	 *            <li>-valid &lt; minValidity &gt; ,i.e., threshold for conf(r)</li>
	 *            <li>-complete &lt; minCompleteness &gt; ,i.e., threshold
	 *            for conf(r bar)</li>
	 *            <li>-query &lt; query &gt; -- to get left outer join of
	 *            dependent and referenced relation, first column must be column
	 *            of referenced relation indicating included (==NOT NULL) or not
	 *            included (==NULL)</li>
	 *            <li>-columns &lt; column names separated by space &gt;</li>
	 *            <li>[-constrained]</li> -- use AprioriConstrained (instead of Apriori)
	 *            <li>[-findCovering]</li> -- find covering conditions (instead of complete)
	 *            <li>[-dontPrintConditions]</li> 
	 *            <li>[-humanReadableOutput]</li> 
	 *            <li>[-suppAbsolute]</li> 
	 *            <li>[-memoryStats]</li> 
	 *            </ul>
	 */
	public static void main(String args[]) 
	throws SQLException, ClassNotFoundException, IOException, ParseException {
		
		
		CindDetectionUsingAprioriExec exec = new CindDetectionUsingAprioriExec();
		
		/*
		 * parse input parameter
		 */
		double minSupport, minValidity, minTotalCompleteness;
		String query;
		String attributes;
		boolean nonConstrained;
		boolean printConditions;
		boolean isSupportAbsolute;
		boolean humanReadableOutput;
		boolean memoryStats;
		boolean findCovering;
		
		try {
			minSupport = exec.getParameterDouble(args, "-supp");
			minValidity = exec.getParameterDouble(args, "-valid");
			minTotalCompleteness = exec.getParameterDouble(args, "-complete");
			query = exec.getParameterString(args, "-query");
			attributes = exec.getParameterString(args, "-columns");
//			numberOfAttributes = exec.getParameterInteger(args, "-nrAttr");
			nonConstrained = !exec.getOption(args, "-constrained");
			printConditions = !exec.getOption(args, "-dontPrintConditions");
			isSupportAbsolute = exec.getOption(args, "-suppAbsolute");
			humanReadableOutput = exec.getOption(args, "-humanReadableOutput");
			memoryStats = exec.getOption(args, "-memoryStats");
			findCovering = exec.getOption(args, "-findCovering");
			
		} catch (ParseException e) {
			exec.printUsage();
			throw e;
		}

		/*
		 * get properties
		 */
		InputStream is = ClassLoader.getSystemResourceAsStream(
				"de/unipotsdam/hpi/cind/associationRuleMining/CindDetectionUsingApriori.properties");
		Properties properties = new Properties();
		properties.load(is);
		
		
		/*
		 * get connection 
		 */
		String hostName = java.net.InetAddress.getLocalHost().getHostName();
		String url;
		if (hostName.equals("anubis.hpi.uni-potsdam.de")) {
			url = properties.getProperty("Database.urlAnubis");
		} else {
			url = properties.getProperty("Database.url");
		}
		Class.forName(properties.getProperty("Database.driver"));
		Connection conn = DriverManager.getConnection(
				url, 
				properties.getProperty("Database.user"), 
				properties.getProperty("Database.password"));
		
		
		
		RuntimeMeasures runtimes = RuntimeMeasures.getInstance();
		runtimes.setMinSupport(minSupport);
		runtimes.setMinValidity(minValidity);
		runtimes.setMinTotalCompleteness(minTotalCompleteness);
		
		
		
		Timer timer = null;
		MemoryCheckerTask task = null;
		
		if (memoryStats) {
			timer = new Timer();
			// start in 2 seconds then run every 0.5 seconds
			task = new MemoryCheckerTask();
			timer.schedule(task, 2000, 500);
		}
		
		long start = System.currentTimeMillis();
		
		/*
		 *  get CINDs
		 */
		CindDetectionUsingApriori detector; 
		if (findCovering) {
			detector = new CoveringCindDetectionUsingApriori(conn);
		} else {
			detector = new CindDetectionUsingApriori(conn);
		}
		
		ArrayList<ConditionalIND> cinds;
		if (nonConstrained) {
			cinds = detector.findConditionalINDsApriori(
					query, attributes.split(" "), minSupport, minValidity, minTotalCompleteness);
		} 
		else if (isSupportAbsolute) {
			cinds = detector.findConditionalINDsAprioriConstrained(
					query, attributes.split(" "), (int) minSupport, minValidity, minTotalCompleteness);
		}
		else
		{
			cinds = detector.findConditionalINDsAprioriConstrained(
					query, attributes.split(" "), minSupport, minValidity, minTotalCompleteness);
		}
		
		conn.close();
		
		long end = System.currentTimeMillis();
		runtimes.setRuntimeOverAll( (int)(end-start) );
		
		
		/*
		 * print CINDs
		 */
		if (printConditions) {
			for (ConditionalIND conditionalIND : cinds) {
				System.out.println(conditionalIND);
			}
		}
		runtimes.setNrCINDs(cinds.size());
		
		
		
		
		if (memoryStats) {
			timer.cancel();
			runtimes.setMaxMemoryUsed( task.getMaxMemoryUsed() );
			runtimes.setAvgMemoryUsed( task.getAverageMemoryUsed() ); 
		}
		
		runtimes.printToStdOut(humanReadableOutput);
		
	}
	
	
    /**
     * Parse input options specifying a String value..
     * 
     * @param args
	 * @param param
     * @return value to specified param
     * @throws ParseException
     */
    private String getParameterString(String[] args, String param)   throws ParseException  {
        try {
            for (int i = 0; i < args.length; i++) {
                String currentArg = args[i];
                if (currentArg.equals(param)) {
                    return args[i+1];
                }
            }
        } catch (IndexOutOfBoundsException e) {
            // not enough parameters
        	throw new ParseException("missing parameter for option " + param, args.length);
        }
        
        throw new ParseException("missing option " + param, args.length);
    }        

	/**
	 * Parse input options specifying a double value.
	 * 
	 * @param args
	 * @param param
	 * @return value to specified param
	 * @throws ParseException
	 */
    private double getParameterDouble(String[] args, String param) throws ParseException {
        try {
            for (int i = 0; i < args.length; i++) {
                String currentArg = args[i];
                if (currentArg.equals(param)) {
                    return Double.parseDouble(args[i+1]);
                }
            }
        } catch (IndexOutOfBoundsException e) {
            // not enough parameters
            throw new ParseException("missing parameter for option " + param, args.length);
        }
        
        throw new ParseException("missing option " + param, args.length);
    }
    
    
    /**
     * Parse input options.
     * 
     * @param args
     * @param option
     * @return true, if option is used
     * @throws ParseException
     */
    private boolean getOption(String[] args, String option) throws ParseException {
    	for (int i = 0; i < args.length; i++) {
    		String currentArg = args[i];
    		if (currentArg.equals(option)) 
    			return true;
    	}
    	return false;
    }
    
    

    /**
     * Print all parameters and options.
     * 
     */
	private void printUsage() {
		
		System.err.println("usage: CindDetectionUsingApriori ");
		System.err.println("	-supp &lt; minSupport &gt;");
		System.err.println("	-valid &lt; minValidity &gt;");
		System.err.println("	-complete &lt; minCompleteness &gt;");
		System.err.println("	-query &lt; query &gt; -- to get left outer join " +
				"of dependent and referenced relation, first column must be column " +
				"of referenced relation indicating included (==NOT NULL) " +
				"or not included (==NULL)");
		System.err.println("	-columns &lt; condition column names separated by space &gt;");
		System.err.println("	[-useConstrained] -- use AprioriConstrained (instead of Apriori)");
		System.err.println("	[-dontPrintConditions]");
		System.err.println("	[-humanReadableOutput]");
		System.err.println("	[-suppAbsolute]");
		System.err.println("	[-memoryStats]");
		
	}
	
}
