/**
 * 
 */
package de.unipotsdam.hpi.cind.associationRuleMining.util;

import java.math.*;
import java.util.*;

/**
 * 
 * HashTree implementation to hold candidate item sets. At level i the i-th item
 * is hashed to determine the next node.
 * 
 * Hash tree containing {@link ItemSetString} objects.
 * 
 * @author jana.bauckmann
 * 
 */
public class HashTreeNode implements Iterable<ItemSetString> {

	public static int DEFAULT_SPLIT_THRESHOLD = 150;
	
	private int splitThreshold;
	
	private HashSet<ItemSetString> leafNode = null;
	private HashMap<String, HashTreeNode> interiorNode = null;
	
	/**
	 * Hash d-th item at depth d.
	 */
	private int depth;
	
	
	private boolean isLeaf;
	
	
	
	/**
	 * Creates an empty HashTreeNode at given level as leaf or interior node.
	 * 
	 * @param depth
	 * @param isLeaf
	 */
	public HashTreeNode(int depth, boolean isLeaf, int splitThreshold) {
		
		this.depth = depth;
		this.isLeaf = isLeaf;
		this.splitThreshold = splitThreshold;
		
		if (isLeaf) {
			this.leafNode = new HashSet<ItemSetString>(splitThreshold * 4/3);
		}
		else {
			this.interiorNode = new HashMap<String, HashTreeNode>(splitThreshold * 4/3);
		}
	}

	/**
	 * Creates an empty HashTreeNode at level 0 as leaf with given split threshold.
	 * 
	 * @param splitThreshold
	 */
	public HashTreeNode(int splitThreshold) {
		
		this(0, true, splitThreshold);
		if (splitThreshold == 0) this.splitThreshold = DEFAULT_SPLIT_THRESHOLD;
	}
		
	/**
	 * Creates an empty HashTreeNode at level 0 as leaf with default split
	 * threshold of HashTreeNode.DEFAULT_SPLIT_THRESHOLD.
	 */
	public HashTreeNode() {
		this(DEFAULT_SPLIT_THRESHOLD);
	}

	
	
	/**
	 * 
	 * @param basket
	 * @return All item sets contained in this HashTreeNode that are a subset of
	 *         given basket.
	 */
	public ArrayList<ItemSetString> getContainedItemSets(Basket basket) {

		ArrayList<ItemSetString> containedSets = new ArrayList<ItemSetString>(100);
		return getContainedItemSets(basket, 0, containedSets);
		
		
	}
	
	private ArrayList<ItemSetString> getContainedItemSets(
			Basket basket, int startIndexBasket, ArrayList<ItemSetString> containedSets) {
		
		if (isLeaf) {
			for (ItemSetString currentSet : leafNode) {
				int itemSetIndex = Math.min(depth, currentSet.size());
				if (basket.contains(currentSet, itemSetIndex, startIndexBasket))
					containedSets.add(currentSet);
			}
			return containedSets;
		
		} else {
		
			HashTreeNode nextNode;
			
			// handle containment of item sets with size < depth
			if (startIndexBasket > 0) {
				nextNode = interiorNode.get(basket.getItem(startIndexBasket-1));
				if (nextNode != null) {
					nextNode.getContainedItemSets(basket, startIndexBasket, containedSets);
				}
			}
			
			
			for (int i = startIndexBasket; i < basket.size(); i++) {
				
				// hash at each item and follow tree
				nextNode = interiorNode.get(basket.getItem(i));
				if (nextNode != null) {
					nextNode.getContainedItemSets(basket, i+1, containedSets);
				}
			}

			return containedSets;
		}
		
	}
	

	
	/**
	 * Increase support of item sets contained in given basket.
	 * 
	 * @param basket
	 * @return true if at least one item set's support was increased
	 */
	public boolean increaseSupportOfContainedItemSets(Basket basket) {
		
		ArrayList<ItemSetString> contained = getContainedItemSets(basket);
		for (ItemSetString itemSet : contained) {
			itemSet.increaseFrequency();
		}
		
		return contained.size() > 0;
	}

	
	/**
	 * 
	 * @param set
	 * @return true if hash tree contains given item set.
	 */
	public boolean contains(ItemSetString set) {
		
		if (isLeaf)
			return leafNode.contains(set);
		else {
			int itemSetIndex = Math.min(depth, set.size()-1);
			HashTreeNode nextNode = (HashTreeNode) interiorNode.get(set.getItem(itemSetIndex));
			if (nextNode == null) 
				return false;
			
			return nextNode.contains(set);
		}
	}
	
	
	/**
	 * Add a new item set.
	 * 
	 * @param set
	 * 
	 * @return true if set was inserted; false if set was already contained.
	 */
	public boolean addItemSet(ItemSetString set) {

		// interior nodes
		// navigate to associated leaf (create it if necessary)
		// and add given set
		if ( !isLeaf ) {
			int itemSetIndex = Math.min(depth, set.size()-1);
			HashTreeNode nextNode = interiorNode.get(set.getItem(itemSetIndex));
			
			if (nextNode != null) {			
				return nextNode.addItemSet(set);
			} else {
				nextNode = new HashTreeNode(depth + 1, true, splitThreshold);
				interiorNode.put(set.getItem(itemSetIndex), nextNode);
				return nextNode.addItemSet(set);
			}
		}
			
		// leaf nodes
		// add the given set if it is not already present
		else  { 

			if (leafNode.contains(set))
				return false; 
			
			if (this.leafNode.size() < splitThreshold) 
				return leafNode.add(set);
			
			this.splitLeaf();
			return this.addItemSet(set);
				

		}
			
		
	}

	/**
	 * Remove item sets from hash tree with frequency lower than specified
	 * supportThreshold.
	 * 
	 * @param supportThreshold
	 * @return true, if at least one item set was removed
	 */
	public boolean pruneInfrequentItemSets(int supportThreshold) {
		
		boolean result = false;
		for (Iterator<ItemSetString> it = this.iterator(); it.hasNext();) {
			ItemSetString currentItemSet = (ItemSetString) it.next();
			if ( currentItemSet.getFrequency() < supportThreshold ) { 
				it.remove();
				result = true;
			}
			
		}
		
		return result;
	}

	/**
	 *  
	 * @param set
	 * @return Frequency of specified <code>set</code>.
	 */
	public int getFrequencyForItemSet(ItemSetString set) {

		if (isLeaf) {
			for (ItemSetString itemSet : leafNode) {
				if (itemSet.equals(set))
					return itemSet.getFrequency();
			}
			return -1;
		}
		else {
			int itemSetIndex = Math.min(depth, set.size()-1);
			HashTreeNode nextNode = (HashTreeNode) interiorNode.get(set.getItem(itemSetIndex));
			if (nextNode == null) 
				return -1;
			
			return nextNode.getFrequencyForItemSet(set);
		}
		
	}
	
	
	/**
	 * 
	 * @return true, if at least one element is contained in this hash tree
	 */
	public boolean isEmpty() {
		
		Iterator<ItemSetString> it = this.iterator();
		return !it.hasNext();
	}
	
	

	/**
	 * Split leaf node.
	 * 
	 * @param set
	 * 
	 */
	private void splitLeaf() {
		
		if ( !isLeaf ) {
			throw new RuntimeException("HashTreeNode.splitLeaf() called at interiorNode");
		}
		
		// create new interiorNode		
		this.interiorNode = new HashMap<String, HashTreeNode>(splitThreshold *4/3);
		
		// copy item sets from old leafNode to new interiorNode.leafNodes
		for (ItemSetString currentSet : leafNode) {

			// subtree exists, add currentSet
			int keyIndex = Math.min(depth, currentSet.size()-1);
			
//			if (keyIndex < depth)
//				System.err.println("itemset " + currentSet.toString() + " too short for depth " + depth);
			
			String key = currentSet.getItem(keyIndex);
			HashTreeNode nextNode = this.interiorNode.get(key);
			if (nextNode != null) {
				nextNode.addItemSet(currentSet);
			} else {
				// create subtree and add currentSet
				nextNode = new HashTreeNode(depth + 1, true, splitThreshold);
				nextNode.addItemSet(currentSet);
				this.interiorNode.put(currentSet.getItem(keyIndex), nextNode);
			}
		}

		isLeaf = false;
		this.leafNode = null;
		
	}
	
	
//	/**
//	 * 
//	 * @return Number of item sets in this node.
//	 */
//	private int size () {
//		if (isLeaf) return leafNode.size();
//		else return interiorNode.size();
//	}

	@Override
	public Iterator<ItemSetString> iterator() {
		
		return new HashTreeIterator(this);
	}
	
	public Iterator<HashSet<ItemSetString>> leafIterator() {
		
		return new HashTreeLeafIterator(this);
		
	}
	
	
	public ArrayList<ItemSetString> toSortedList() {
		ArrayList<ItemSetString> result = new ArrayList<ItemSetString>(splitThreshold * 5);
		
		for (Iterator<HashSet<ItemSetString>> it = this.leafIterator(); it.hasNext();) {
			result.addAll(it.next());
		}
		Collections.sort(result);
		
		return result;
		
	}
	
	
	/**
	 * 
	 * @return Number of item sets in this hash tree.
	 */
	public int size() {
		
		int size = 0;
		for (Iterator<HashSet<ItemSetString>> it = this.leafIterator(); it.hasNext();) {
			size += it.next().size();
		}
		
		return size;
	}
	
	/**
	 *  
	 * @param set
	 * @return Depth of leaf containing <code>set</code>.
	 */
	public int getDepthForItemSet(ItemSetString set) {

		if (isLeaf) {
			if (!leafNode.contains(set))
				throw new RuntimeException("tree does not contain given item set");
		
			return depth;
		}
		else {
			int itemSetIndex = Math.min(depth, set.size()-1);
			HashTreeNode nextNode = (HashTreeNode) interiorNode.get(set.getItem(itemSetIndex));
			if (nextNode == null)
				throw new RuntimeException("tree does not contain given item set");
			
			return nextNode.getDepthForItemSet(set);
		}
	}
	
	/**
	 * 
	 * @param set
	 * @return Size of leaf containing <code>set</code>.
	 */
	public int sizeOfLeafContaining(ItemSetString set) {

		if (isLeaf) { 
			if (!leafNode.contains(set))
				throw new RuntimeException("tree does not contain given item set");
		
			return leafNode.size();
		}
		else {
			int itemSetIndex = Math.min(depth, set.size()-1);
			HashTreeNode nextNode = (HashTreeNode) interiorNode.get(set.getItem(itemSetIndex));
			if (nextNode == null)
				throw new RuntimeException("tree does not contain given item set");
			
			return nextNode.sizeOfLeafContaining(set);
		}
	}
	
	public String toString() {
		
		StringBuffer sb = new StringBuffer();
		String blank = "";
		for (int i = 0; i < depth; i++) {
			blank += "	";	
		}
		
		
		if (isLeaf) {
			
			sb.append(blank).append("leaf: ");
			for (ItemSetString currentItemSet : leafNode) {
				sb.append("(")
					.append(currentItemSet.toString())
					.append("), ");
			}
			sb.append("\n");
			
		} else {
			
			for (String currentKey : interiorNode.keySet()) {
				sb.append(blank)
					.append("interiorNode[");
				sb.append(currentKey);
				sb.append("] \n");
				sb.append((interiorNode.get(currentKey)).toString());
			}
			
		}
		
		
		
		return sb.toString();
		
	}
	
	public String toStringShort() {
		
		StringBuffer sb = new StringBuffer();
		String blank = "";
		for (int i = 0; i < depth; i++) {
			blank += "	";	
		}
		
		
		sb.append(blank);
		if (isLeaf) {
			
			if (leafNode.isEmpty()) {
				sb.append("EMPTY leaf");
			}
			else {
				sb.append("leaf with ");
				sb.append(leafNode.size());
				sb.append(" item sets \n");
			}
			
		} else {
			
			for (String currentKey : interiorNode.keySet()) {
				sb.append(blank)
					.append("interiorNode[");
				sb.append(currentKey);
				sb.append("] \n");
				sb.append((interiorNode.get(currentKey)).toStringShort());
			}
			
		}
		
		
		
		return sb.toString();
		
	}
	
	private class HashTreeIterator implements Iterator<ItemSetString> {

		private Deque<Iterator<HashTreeNode>> nodeIterators;
		private Iterator<ItemSetString> leafIterator;
		
		HashTreeIterator(HashTreeNode rootNode) {
			
			nodeIterators = new ArrayDeque<Iterator<HashTreeNode>>();
			traverseToNextLeaf(rootNode);
			
		}
		
		

		
		
		@Override
		public boolean hasNext() {
			
			if (leafIterator.hasNext()) 
				return true;
			
			/*
			 * traverse to next leaf
			 */
			// traverse up till next non-finished iterator
			if ( nodeIterators.isEmpty() ) // handle hash trees with one leaf only
				return false;
			Iterator<HashTreeNode> currentNodeIterator = nodeIterators.pop();
			while ( !currentNodeIterator.hasNext() ) {
				if ( nodeIterators.isEmpty() ) // all interior nodes iterated
					return false;
				currentNodeIterator = nodeIterators.pop();
			}
			nodeIterators.push(currentNodeIterator);
			// traverse down to next leaf
			traverseToNextLeaf(currentNodeIterator.next());
			
			
			/*
			 *  call hasNext() to handle empty leafs
			 */
			return hasNext();
		}

		@Override
		public ItemSetString next() {

			if (leafIterator.hasNext()) 
				return leafIterator.next();
			
			/*
			 * traverse to next leaf
			 */
			// traverse up till next non-finished iterator
			if ( nodeIterators.isEmpty() ) // handle hash trees with one leaf only
				throw new NoSuchElementException();
			Iterator<HashTreeNode> currentNodeIterator = nodeIterators.pop();
			while ( !currentNodeIterator.hasNext() ) {
				if (nodeIterators.isEmpty()) // all interior nodes iterated
					throw new NoSuchElementException();
				currentNodeIterator = nodeIterators.pop();
			}
			nodeIterators.push(currentNodeIterator);
			// traverse down to next leaf
			traverseToNextLeaf(currentNodeIterator.next());
			
			/*
			 *  return next item set 
			 *  (call next() to handle empty leafs)
			 */
			return this.next();
			
		}

		
		
		private void traverseToNextLeaf(HashTreeNode startNode) {

			HashTreeNode currentNode = startNode;
			Iterator<HashTreeNode> currentNodeIterator;
			
			// process interior nodes
			while ( !currentNode.isLeaf ) {
				currentNodeIterator = currentNode.interiorNode.values().iterator();
				nodeIterators.push(currentNodeIterator);
				
				if (currentNodeIterator.hasNext()) {
					currentNode = currentNodeIterator.next();
				}
				else throw new IllegalStateException();
			}
			
			// found leaf node
			leafIterator = currentNode.leafNode.iterator();
		}
		
		@Override
		public void remove() {
			leafIterator.remove();
			
		}
		
	}
	
	
	
	private class HashTreeLeafIterator implements Iterator<HashSet<ItemSetString>> {

		private Deque<Iterator<HashTreeNode>> nodeIterators;
		HashSet<ItemSetString> leaf;
		
		HashTreeLeafIterator(HashTreeNode rootNode) {
			
			nodeIterators = new ArrayDeque<Iterator<HashTreeNode>>();
			leaf = traverseToNextLeaf(rootNode);
			try {
				while (leaf.isEmpty()) {
					leaf = findNext();
				}
			} catch (NoSuchElementException e) {
				leaf = null;
			}
			
		}
		
		

		
		
		@Override
		public boolean hasNext() {
			
			return this.leaf != null && !this.leaf.isEmpty();
		}

		@Override
		public HashSet<ItemSetString> next() {
			
			HashSet<ItemSetString> next = this.leaf;
			
			
			try {
				do {
					leaf = findNext();
				} while (leaf.isEmpty());
			} catch (NoSuchElementException e) {
				leaf = null;
			}
			
			return next;
		}
		
		
		private HashSet<ItemSetString> findNext() {

			/*
			 * traverse to next leaf
			 */
			
			HashSet<ItemSetString> leaf;
			do {
				// traverse up till next non-finished iterator
				Iterator<HashTreeNode> currentNodeIterator = nodeIterators.pop();
				while ( !currentNodeIterator.hasNext() ) {
					if (nodeIterators.isEmpty())
						throw new NoSuchElementException();
					currentNodeIterator = nodeIterators.pop();
				}
				nodeIterators.push(currentNodeIterator);
				// traverse down to next leaf
				leaf = traverseToNextLeaf(currentNodeIterator.next());
			} while (leaf.isEmpty()); 
			
			return leaf;
			
		}

		
		
		private HashSet<ItemSetString> traverseToNextLeaf(HashTreeNode startNode) {

			HashTreeNode currentNode = startNode;
			Iterator<HashTreeNode> currentNodeIterator;
			
			// process interior nodes
			while ( !currentNode.isLeaf ) {
				currentNodeIterator = currentNode.interiorNode.values().iterator();
				nodeIterators.push(currentNodeIterator);
				
				if (currentNodeIterator.hasNext()) {
					currentNode = currentNodeIterator.next();
				}
				else throw new IllegalStateException();
			}
			
			// found leaf node
			return currentNode.leafNode;
		}
		
		@Override
		public void remove() {
			throw new UnsupportedOperationException();
			
		}
		
	}
	
	
}
