package de.unipotsdam.hpi.cind.associationRuleMining.util;

import java.util.*;


/**
 * Represent a k-item-set and its support count.
 * 
 * @author jana.bauckmann
 *
 */
public class ItemSetString implements Iterable<String>, Comparable<ItemSetString> {

	private ArrayList<String> items;
	
	private int supportCount;


	
	/**
	 * New item set with specified <code>items</code>. Order of items is preserved.
	 * 
	 * @param items
	 */
	public ItemSetString(List<String> items) {
		this.items = new ArrayList<String>(items);
		this.supportCount = 0;
		
	}

	/**
	 * New item set with specified <code>items</code>. Order of items is preserved.
	 * 
	 * @param items
	 */
	public ItemSetString(String... items) {
		this(Arrays.asList(items));
	}

	
	/**
	 * New item set starting with specified <code>items</code> followed by
	 * specified <code>lastItem</code>. Order of items is preserved.
	 * 
	 * @param items
	 * @param lastItem
	 */
	public ItemSetString(List<String> items, String lastItem) {
		this.items = new ArrayList<String>(items);
		this.items.add(lastItem);
		this.supportCount = 0;
	}
	
	/**
	 * New item set starting with specified <code>firstItems</code> followed by
	 * specified <code>lastItems</code>. Order of items is preserved.
	 * 
	 * @param items
	 * @param lastItem
	 */
	public ItemSetString(List<String> firstItems, List<String> lastItems) {
		this.items = new ArrayList<String>(firstItems.size() + lastItems.size());
		this.items.addAll(firstItems);
		this.items.addAll(lastItems);
		this.supportCount = 0;
	}
	
	
	public void increaseFrequency() {
		this.supportCount++;
	}
	
	public int getFrequency() {
		return supportCount;
	}
	
	public String getItem(int index) {
		return items.get(index);
	}
	
	public ArrayList<String> getItems() {
		return items;
	}
	
	/**
	 * 
	 * @return Number of items in this item set.
	 */
	public int size() {
		return items.size();
	}
	
	@Override
	public boolean equals(Object o) {
		
        if ( !(o instanceof ItemSetString) )
            return false;
        
        ItemSetString other = (ItemSetString) o;
        
        if ( this.items == null 
                ? other.items != null 
                : !this.items.equals(other.items) )
            return false;
        
        
        
        // passed all tests
        return true;
	}
	
	@Override
	public int hashCode() {
		
		return items.hashCode();
	}
	

	/**
	 * Compares item lists. First unequal items determine the order, i.e., 
	 * (foo, bar, z) < (foo, blub, a).
	 * 
	 * @param other
	 * @return negative integer, zero, or a positive integer as this object is
	 *         less than, equal to, or greater than the specified object
	 */
	@Override
	public int compareTo(ItemSetString other) {

		int result;

		// compare overlapping parts of items lists 
		for (int i = 0; i < Math.min(items.size(), other.items.size()); i++) {
			result = items.get(i).compareTo(other.items.get(i));
			if (result != 0) 
				return result;
		}
		
		// all items equal so far...
		// equal sized item sets (general case)
		if (items.size() == other.items.size())
			return 0;
		
		// more items in other item set
		if (items.size() < other.items.size())
			return -1;
		
		// more items in this item set
		return 1;
		
		
	}

	
	@Override
	public Iterator<String> iterator() {
		return this.items.iterator();
	}
	
	/**
	 * Returns a list iterator of the elements in this item set (in proper
	 * sequence), starting at the specified position. The specified
	 * index indicates the first element that would be returned by an initial
	 * call to next. An initial call to previous would return the element with
	 * the specified index minus one.
	 * 
	 * @param index
	 *            index of first element to be returned from the list iterator
	 *            (by a call to the next method)
	 * 
	 * @return a list iterator of the elements in this list (in proper
	 *         sequence), starting at the specified position in this list
	 *         
	 * 
	 */
	public Iterator<String> listIterator(int index) {
		return this.items.listIterator(index);
	}
	
	@Override
	public String toString() {
		
		StringBuffer sb = new StringBuffer();
		sb.append(items.toString());
		sb.append("[").append(supportCount).append("x]");
		return sb.toString();
	}


	
}
