package de.unipotsdam.hpi.cind.unaryCondition;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;


/**
 * 
 * @author benjamin.emde
 *
 */
public class CIndFinder {

	/**
	 *  CIND detection mode
	 */
	static enum V{COMPLETE, TOTALLY};
	
	/**
	 * Find all (totally) complete CINDs for given embedded IND. SQL statements
	 * are used to check CINDs. Ask table ALADIN.colum_infos for condition
	 * attributes.
	 * 
	 * * @param args
	 * <ul>
	 * <li>usage: dep ref [complete|totally] maxConflicts]</li>
	 * <li>dep dependent attribute of embedded IND (schema.table.column)</li>
	 * <li>ref referenced attribute of embedded IND (schema.table.column)</li>
	 * <li>[complete|totally] find/test? complete or totally complete CINDs</li>
	 * <li>maxConflicts total? number of maximally allowed conflicts</li>
	 * </ul>
	 */
	public static void main(String[] args) {
		if (args.length < 4) {
			System.err.println("Usage: dep ref [complete|totally] maxConflicts");
			return;
		}
		
		// parse arguments
		String[] parts = args[0].split("\\.");
		String depSchema = parts[0];
		String depTable = parts[1];
		String depCol = parts[2];
		parts = args[1].split("\\.");
		String refSchema = parts[0];
		String refTable = parts[1];
		String refCol = parts[2];
		
		V version = args[2].equals("totally") ? V.TOTALLY : V.COMPLETE;
		int maxConflicts = Integer.parseInt(args[3]);
		
		// call CIND detection
		identifyCInd(depSchema, depTable, depCol,
				refSchema, refTable, refCol,
				version, maxConflicts);
	}

	/**
	 * 
	 * @param depSchema
	 * @param depTable
	 * @param depCol
	 * @param refSchema
	 * @param refTable
	 * @param refCol
	 * @param version
	 * @param maxConflicts
	 * @return
	 */
	public static int identifyCInd(String depSchema, String depTable, String depCol,
			String refSchema, String refTable, String refCol,
			V version, int maxConflicts) {
		int count = 0;
		
		try {

			
			// fetch condition-column candidates
			long start = System.currentTimeMillis();
			List<String> condCols = queryCondCols(depSchema, depTable, depCol);
			
			if(condCols.size() == 0)
				return 0;

			// fetch statistics, i.e., number of tuples and number of distinct
			// tuples in dependent column
			ResultSet rsCount = Db.getInstance().con
					.createStatement()
					.executeQuery(
							buildCountQuery(depSchema + "." + depTable, depCol));
			rsCount.next();
			ResultSet joinResult;
			
			// iterate over condition-column candidates
			for (String condCol : condCols) {
				joinResult = Db.getInstance().con.createStatement().executeQuery(
						buildJoinQuery(refSchema + "." + refTable, refCol,
								depSchema + "." + depTable, depCol, condCol));
				
				// call totally complete CIND detection
				if (version.equals(V.TOTALLY)) {

					Object tCond = identifyConditionForTotallyCompleteCIND(joinResult, condCol, maxConflicts);
					
					if(tCond != null) {
						System.out.println("Valid & Totally Complete Conditional INDs: " + condCol + " = \"" + tCond + "\"");
						count++;
					}
				
				// call complete CIND detection
				} else if (version.equals(V.COMPLETE)) {

					Set<Object> gConds = identifyConditionsForCompleteCIND(joinResult, condCol, maxConflicts);
					
					for (Object cond : gConds) {
						System.out.println("Valid & Complete Conditional INDs: " + condCol + " = \"" + cond + "\"");
					}
					count += gConds.size();
					
				}
				joinResult.close();
			}
			System.out.println(condCols.size() + ", " + rsCount.getInt("AnzahlTupel") + ", " + rsCount.getInt("AnzahlB") + ", " + count + ", " + (System.currentTimeMillis() - start));
			
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return count;
	}
	
	/**
	 * Get condition over condCol for totally complete CIND.
	 * 
	 * Prints conflicting tuples to standard out and returns condition.
	 * 
	 * @param joinResult
	 * @param condCol
	 * @return valid & totally complete tCond for <code>condCol</code>
	 * @throws SQLException
	 */
	private static Object identifyConditionForTotallyCompleteCIND(ResultSet joinResult, String condCol, int maxConflicts)
			throws SQLException {
		
		List<ConflictingTuple> conflictingTuples = new LinkedList<ConflictingTuple>();
		
		// available condition (as known so far)
		Object tCond = null;
		// first available condition found and set in tCond
		boolean tCondInitialized = false;
		// set of unavailable conditions, 
		// i.e., found a tuple with b != c and this value in a 
		Set<Object> unavailableConditions = new HashSet<Object>();
		
		// value of conditional attribute
		Object a; 
		// value of referenced attribute
		Object c;
		// value of dependent attribute
		Object b;
		
		while (joinResult.next() == true) {
			b = joinResult.getObject(1);
			if (b == null)	// couldn't distinguish referring vs. non-referring (c==null in both cases)
				continue;
			a = joinResult.getObject(2);
			c = joinResult.getObject(3);
			
			// b not in C (dependent value not in referenced attribute)
			// i.e., value in "a" is no available condition  
			if(c == null) {
				if(!tCondInitialized) {
					unavailableConditions.add(a);
				} else if ((tCond == null && a == null) || tCond.equals(a)) {
					// not available anymore
					conflictingTuples.add(new ConflictingTuple(condCol, a, b, c) );
					if(conflictingTuples.size() > maxConflicts) {
						return null;
					}
				}
			// b in C (dependent value in referenced attribute)
			// i.e., value in "a" is an available condition (as known so far)
			} else {
				if(!tCondInitialized) {
					if(unavailableConditions.contains(a))
						return null;	// not totally complete
					tCond = a;
					tCondInitialized = true;
				} else if((tCond == null && a != null) || !tCond.equals(a)) {
					return null;	// not totally complete anymore
				}
			}
		}
		joinResult.close();

		for (ConflictingTuple conflictingTuple : conflictingTuples) {
			System.out.println(conflictingTuple.toString());
		}

		if(tCondInitialized && tCond == null)
			return "NULL";
		
		return tCond;
	}

	/**
	 * Get conditions over condCol for  complete CIND.
	 * 
	 * @param joinResult
	 * @param condCol
	 * @return set of valid & complete gConds for <code>condCol</code>
	 * @throws SQLException
	 */
	private static Set<Object> identifyConditionsForCompleteCIND(ResultSet joinResult, String condCol, int maxConflicts) throws SQLException {

		// available condition (as known so far)
		Set<Object> gConds = new HashSet<Object>();
		// set of unavailable conditions, 
		// i.e., found a tuple with b != c and this value in a 
		Set<Object> unavailableConditions = new HashSet<Object>();
		Set<Object> aValues = new HashSet<Object>();
		
		List<ConflictingTuple> conflictingTuples = new LinkedList<ConflictingTuple>();
		

		// value of conditional attribute
		Object a;
		// value of dependent attribute
		Object b = null;
		// value of referenced attribute
		Object c = null;
		
		
		/*
		 * Get groups over b and collect values in a in aValues.
		 * Once a group is completed, update condition candidates.
		 */
		while (joinResult.next() == true) {
			a = joinResult.getObject(2);
			if (aValues.isEmpty() || b == null) {
				b = joinResult.getObject(1);
				if (b == null)	// couldn't distinguish referring vs. non-referring (c==null in both cases)
					continue;
				c = joinResult.getObject(3);
			} else if (! b.equals(joinResult.getObject(1))) {	// set of a-values complete for one b
				if( !updateConditionCandidates(gConds, aValues, c, unavailableConditions) ) {
					// TODO: make conflict output more specific
					conflictingTuples.add(new ConflictingTuple(condCol, Arrays.toString(aValues.toArray()), b, c) );
					if(conflictingTuples.size() > maxConflicts) {
						return Collections.emptySet();
					}
				}
				aValues.clear();
				b = joinResult.getObject(1);
				if (b == null)	// couldn't distinguish referring vs. non-referring (c==null in both cases)
					continue;
				c = joinResult.getObject(3);
			}
			aValues.add(a);
		}

		if( !updateConditionCandidates(gConds, aValues, c, unavailableConditions) ) {	// set of a-values complete for last b
			// TODO: make conflict output more specific
			conflictingTuples.add(new ConflictingTuple(condCol, Arrays.toString(aValues.toArray()), b, c) );
			if(conflictingTuples.size() > maxConflicts) {
				return Collections.emptySet();
			}
		}

//		for (ConflictingTuple conflictingTuple : conflictingTuples) {
//			System.out.println(conflictingTuple.toString());
//		}
		
		joinResult.close();

		if (gConds.remove(null))
			gConds.add("NULL");
		
		return gConds;
	}

	/**
	 * Build following statement: 
	 * 
	 * select d."depCol", d."condCol", MIN(r."refCol") 
	 * from depSchemaTable d left outer join refSchemaTable r on d."depCol" = r."refCol" 
	 * group by d."depCol", d."condCol"
	 * 
	 * select B, A, MIN(C) 
	 * from R left outer join S on B = C 
	 * group by B, A
	 * 
	 * 
	 * @param refSchemaTable referenced table (with schema) of embedded IND
	 * @param refCol referenced column of embedded IND
	 * @param depSchemaTable dependent table (with schema) of embedded IND
	 * @param depCol dependent column of embedded IND
	 * @param condCol possible condition column
	 * @return
	 */
	private static String buildJoinQuery(String refSchemaTable, String refCol,
			String depSchemaTable, String depCol, String condCol) {
		StringBuilder b = new StringBuilder();
		b.append("select d.\"").append(depCol).append("\", d.\"").append(condCol)
				.append("\", MIN(r.\"").append(refCol).append("\") ");
		b.append("from ").append(depSchemaTable).append(" d left outer join ")
				.append(refSchemaTable).append(" r ");
		b.append("on d.\"").append(depCol).append("\" = r.\"").append(refCol).append(
				"\" ");
		b.append("group by d.\"").append(depCol).append("\", d.\"").append(condCol)
				.append("\"");
		return b.toString();
	}

	/**
	 * This method updates the condCands and invalidConditions set for one condition-column.
	 * 
	 * @param gConds			the set of all gCond-candidates that has to be updated
	 * @param aValues				the condition-column values used for existence- or absence-checking w/ <code>gConds</code>
	 * @param c						referenced PK or null
	 * @param invalidConditions		invalid conditions so far
	 * @return <code>false</code> iff no valid condition left
	 */
	static boolean updateConditionCandidates(Set<Object> gConds,
			Set<Object> aValues, Object c, Set<Object> invalidConditions) {
		
		if (c != null) {
			// t[B] element C
			if (gConds.isEmpty()) {
				// initial fill of condition-candidate set
				gConds.addAll(aValues);
				gConds.removeAll(invalidConditions);
				invalidConditions.clear();
				if (gConds.isEmpty())
					return false;
			} else {
				// intersect the conditions
				gConds.retainAll(aValues);
				if (gConds.isEmpty())
					return false;
			}
		} else {
			// t[B] not element C
			if (gConds.isEmpty()) {
				// save invalid conditions until the condition-candidate set is filled initially
				invalidConditions.addAll(aValues);
			} else {
				// filter invalid conditions from all valid conditions
				gConds.removeAll(aValues);
				if (gConds.isEmpty())
					return false;
			}
		}
		
		return true;
	}

	/**
	 * Ask table ALADIN.colum_infos for condition attributes (i.e. all
	 * attributes in depTable except depCol).
	 * 
	 * @param depSchema
	 *            dependent schema of embedded IND
	 * @param depTable
	 *            dependent table of embedded IND
	 * @param depCol
	 *            dependent column of embedded IND
	 * @return {@link java.util.List} with condition attributes (i.e. column
	 *         names as {@link String})
	 * @throws SQLException
	 */
	static List<String> queryCondCols(String depSchema,
			String depTable, String depCol) throws SQLException {

		List<String> condCols = new LinkedList<String>();
		
		String q = "select COLUMN_NAME " + "from ALADIN.COLUMN_INFOS "
				+ "where SCHEMA_NAME = \'" + depSchema + "\' "
				+ "and TABLE_NAME = \'" + depTable + "\' "
				+ "and COLUMN_NAME not like \'" + depCol + "\' "
				+ "and DATA_TYPE not like \'CLOB\' "
				+ "and NUM_DISTINCT_COLUMN > 1";
		
		ResultSet rs = Db.getInstance().con.createStatement().executeQuery(q);
		while (rs.next() == true) {
			condCols.add(rs.getString("COLUMN_NAME"));
		}
		rs.close();
		
		return condCols;
	}
	
	/**
	 * Build SQL statement to query number of tuples and number of distinct
	 * tuples in depSchemaTable.depCol
	 * 
	 * @param depSchemaTable
	 * @param depCol
	 * @return
	 */
	static String buildCountQuery(String depSchemaTable, String depCol) {
		StringBuilder b = new StringBuilder();
		b.append("select count(*) as AnzahlTupel, count(distinct \"").append(depCol).append("\") as AnzahlB ");
		b.append("from ").append(depSchemaTable);
		return b.toString();
	}
	
	public static Collection<String> queryTestcases() throws SQLException,
			InstantiationException, IllegalAccessException,
			ClassNotFoundException {
		Set<String> testCases = new HashSet<String>();
		String q = "SELECT * FROM ALADIN.INCL_DEPS " +
				"where ALLOWED_DIRTY_DATA_RATE != 0 or ALLOWED_DIRTY_DATA_RATE is null";
		
		ResultSet rs = Db.getInstance().con.createStatement().executeQuery(q);
		while (rs.next() == true) {
			testCases.add(rs.getString(2) + "." + rs.getString(3) + "." + rs.getString(4)
					+ " " + rs.getString(5) + "." + rs.getString(6) + "." + rs.getString(7));
		}
		rs.close();
		return testCases;
	}
}
